#include <stdio.h>
#include <string.h>
#ifdef _WIN32
#include <windows.h>
#include <windowsx.h>
#include <objbase.h>
#include <initguid.h>
#include <commctrl.h>
#else
#include <gtk/gtk.h>
#endif
#include <SDL/SDL.h>
#include "plugin.h"
#include "dspad.h"
#include "resource.h"

#pragma comment(lib, "sdl.lib")
#pragma comment(lib, "comctl32.lib")

#define PLUGIN_NAME "DsPad plugin 1.6.1"
DEFINE_GUID(PLUGGUID, 0x8229CDC4, 0x89AB, 0x45CF, 0xBA, 0x01, 0x3F, 0xB8, 0x7A, 0xD9, 0x17, 0xE9);

#ifdef _WIN32
BOOL CfgDlgProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam);

extern HINSTANCE hInst;
#else
void on_btn_clicked(GtkButton *button);
void on_axisthresh_changed(GtkComboBox *combo);
void on_saveconf_clicked();
#endif
DsPadClass DsPad;
SDL_Joystick *joy;
int btn[8];
int axis_thresh;

void Message(char *txt, ...)
{
	va_list list;
	char msg[512];
	va_start(list, txt);
	vsprintf(msg, txt, list);
	va_end(list);
	#ifdef _WIN32
	MessageBox(0, msg, "DsPad plugin Msg", MB_ICONINFORMATION);
	#else
	GtkWidget *msgBox=gtk_message_dialog_new(NULL, GTK_DIALOG_MODAL, GTK_MESSAGE_INFO, GTK_BUTTONS_OK, "%s", msg);
	gtk_dialog_run(GTK_DIALOG(msgBox));
	gtk_widget_destroy(msgBox);
	#endif
}

void SetButtonText(void *button, char *txt, ...)
{
	va_list list;
	char msg[512];
	va_start(list, txt);
	vsprintf(msg, txt, list);
	va_end(list);
	#ifdef _WIN32
	SetWindowText((HWND)button, msg);
	#else
	gtk_button_set_label(GTK_BUTTON(button), msg);
	#endif
}

void AddComboItem(void *combo, char *txt, ...)
{
	va_list list;
	char msg[512];
	va_start(list, txt);
	vsprintf(msg, txt, list);
	va_end(list);
	#ifdef _WIN32
	ComboBox_AddString((HWND)combo, msg);
	#endif
}

DsPadClass::DsPadClass()
{
}

DsPadClass::~DsPadClass()
{
}

void DsPadClass::MakeCfgFileName(char *fName)
{
	#ifdef _WIN32
	GetModuleFileName(hInst, fName, 256);
	strcpy(fName+strlen(fName)-3, "ini");
	#else
	strcpy(fName, "DsPad.ini");
	#endif
}

void DsPadClass::SaveConf()
{
	char fName[256];
	DsPad.MakeCfgFileName(fName);
	FILE *cf=fopen(fName, "w");
	if(cf==NULL){Message("Error : could not open config file"); return;}
	for(int i=0; i<=7; i++)
	{
		fprintf(cf, "Btn%i=%i\n", i, btn[i]);
	}
	fprintf(cf, "Axis threshold=%i\n", axis_thresh);
	fclose(cf);
}

void DsPadClass::LoadConf()
{
	char fName[256];
	int a=0;
	axis_thresh=1000;
	DsPad.MakeCfgFileName(fName);
	FILE *cf=fopen(fName, "r");
	if(cf==NULL){DsPad.SaveConf(); return;}
	for(int i=0; i<=7; i++)
	{
		fscanf(cf, "Btn%i=%i\n", &a, &btn[i]);
	}
	fscanf(cf, "Axis threshold=%i\n", &axis_thresh);
	fclose(cf);
}

int DsPadClass::Init()
{
	SDL_Init(SDL_INIT_JOYSTICK);
	joy=SDL_JoystickOpen(0);
	if(!SDL_JoystickOpened(0))Message("Error %i opening the joystick", *SDL_GetError());
	DsPad.LoadConf();
	return TRUE;
}

int DsPadClass::Quit()
{
	if(SDL_JoystickOpened(0))SDL_JoystickClose(joy);
	SDL_Quit();
	return TRUE;
}

DWORD DsPadClass::GetInfo(LPGETPLUGININFO info)
{
	info->dwType=PIT_PAD;
	if(info->pszText!=NULL)strcpy((char *)info->pszText, PLUGIN_NAME);
	info->guidID=PLUGGUID;
	return TRUE;
}

BOOL DsPadClass::SetInfo(LPSETPLUGININFO info)
{
	if(info==NULL)return FALSE;
	if(PLUGINISENABLE(info))return DsPad.Init();
   	return TRUE;
}

BOOL DsPadClass::SetProperty(LPSETPROPPLUGIN prop)
{
	#ifdef _WIN32
	PROPSHEETPAGE Psp={NULL};
	HPROPSHEETPAGE hPsp;
	Psp.dwSize=sizeof(PROPSHEETPAGE);
	Psp.dwFlags=PSP_DEFAULT|PSP_USETITLE;
	Psp.hInstance=hInst;
	Psp.pfnDlgProc=(DLGPROC)CfgDlgProc;
	Psp.pszTemplate=MAKEINTRESOURCE(IDD_CFG);
	Psp.pszTitle="DsPad Plugin Config";
	hPsp=CreatePropertySheetPage(&Psp);
	if(hPsp==NULL)return FALSE;
	return PropSheet_AddPage((HWND)prop->hwndOwner, hPsp);
	#else
	GtkWidget *frame=gtk_fixed_new(), *label=gtk_label_new("DsPad Plugin Config");
	GtkWidget *btn1, *btn2, *btn3, *btn4, *btn5, *btn6, *btn7, *btn8;
	GtkWidget *label1;
	GtkWidget *_axis_thresh;
	GtkWidget *saveConf;
	gtk_widget_show(frame);
	gtk_widget_show(label);
	btn1=gtk_button_new_with_label("A : ");
	btn2=gtk_button_new_with_label("B : ");
	btn3=gtk_button_new_with_label("Select : ");
	btn4=gtk_button_new_with_label("Start : ");
	btn5=gtk_button_new_with_label("L : ");
	btn6=gtk_button_new_with_label("R : ");
	btn7=gtk_button_new_with_label("X : ");
	btn8=gtk_button_new_with_label("Y : ");
	gtk_fixed_put(GTK_FIXED(frame), btn1, 8, 8+(2*24));
	gtk_fixed_put(GTK_FIXED(frame), btn2, 8, 8+(3*24));
	gtk_fixed_put(GTK_FIXED(frame), btn3, 8, 8+24);
	gtk_fixed_put(GTK_FIXED(frame), btn4, 152, 8+24);
	gtk_fixed_put(GTK_FIXED(frame), btn5, 8, 8);
	gtk_fixed_put(GTK_FIXED(frame), btn6, 152, 8);
	gtk_fixed_put(GTK_FIXED(frame), btn7, 152, 8+(2*24));
	gtk_fixed_put(GTK_FIXED(frame), btn8, 152, 8+(3*24));
	gtk_widget_set_size_request (btn1, 144, 24);
	gtk_widget_set_size_request (btn2, 144, 24);
	gtk_widget_set_size_request (btn3, 144, 24);
	gtk_widget_set_size_request (btn4, 144, 24);
	gtk_widget_set_size_request (btn5, 144, 24);
	gtk_widget_set_size_request (btn6, 144, 24);
	gtk_widget_set_size_request (btn7, 144, 24);
	gtk_widget_set_size_request (btn8, 144, 24);
	gtk_widget_set_name(btn1, "btn1");
	gtk_widget_set_name(btn2, "btn2");
	gtk_widget_set_name(btn3, "btn3");
	gtk_widget_set_name(btn4, "btn4");
	gtk_widget_set_name(btn5, "btn5");
	gtk_widget_set_name(btn6, "btn6");
	gtk_widget_set_name(btn7, "btn7");
	gtk_widget_set_name(btn8, "btn8");
	gtk_widget_show(btn1);
	gtk_widget_show(btn2);
	gtk_widget_show(btn3);
	gtk_widget_show(btn4);
	gtk_widget_show(btn5);
	gtk_widget_show(btn6);
	gtk_widget_show(btn7);
	gtk_widget_show(btn8);
	g_signal_connect((gpointer)btn1, "clicked", G_CALLBACK(on_btn_clicked), NULL);
	g_signal_connect((gpointer)btn2, "clicked", G_CALLBACK(on_btn_clicked), NULL);
	g_signal_connect((gpointer)btn3, "clicked", G_CALLBACK(on_btn_clicked), NULL);
	g_signal_connect((gpointer)btn4, "clicked", G_CALLBACK(on_btn_clicked), NULL);
	g_signal_connect((gpointer)btn5, "clicked", G_CALLBACK(on_btn_clicked), NULL);
	g_signal_connect((gpointer)btn6, "clicked", G_CALLBACK(on_btn_clicked), NULL);
	g_signal_connect((gpointer)btn7, "clicked", G_CALLBACK(on_btn_clicked), NULL);
	g_signal_connect((gpointer)btn8, "clicked", G_CALLBACK(on_btn_clicked), NULL);
	label1=gtk_label_new("Axis sensitivity :");
	gtk_fixed_put(GTK_FIXED(frame), label1, 56, 112);
	gtk_widget_set_size_request(label1, 104, 24);
	gtk_widget_show(label1);
	_axis_thresh=gtk_combo_box_new_text();
	gtk_fixed_put(GTK_FIXED(frame), _axis_thresh, 160, 112);
	gtk_widget_set_size_request(_axis_thresh, 88, 24);
	gtk_widget_show(_axis_thresh);
	g_signal_connect((gpointer)_axis_thresh, "changed", G_CALLBACK(on_axisthresh_changed), NULL);
	saveConf=gtk_button_new_with_label("Save config");
	gtk_fixed_put(GTK_FIXED(frame), saveConf, 88, 160);
	gtk_widget_set_size_request(saveConf, 144, 24);
	gtk_widget_show(saveConf);
	g_signal_connect((gpointer)saveConf, "clicked", G_CALLBACK(on_saveconf_clicked), NULL);
	gtk_notebook_append_page(GTK_NOTEBOOK(prop->hwndOwner), frame, label);
	if(btn[0])SetButtonText(btn1, "A : Button %i", btn[0]);
	else SetButtonText(btn1, "A : None");
	if(btn[1])SetButtonText(btn2, "B : Button %i", btn[1]);
	else SetButtonText(btn2, "B : None");
	if(btn[2])SetButtonText(btn3, "Select : Button %i", btn[2]);
	else SetButtonText(btn3, "Select : None");
	if(btn[3])SetButtonText(btn4, "Start : Button %i", btn[3]);
	else SetButtonText(btn4, "Start : None");
	if(btn[4])SetButtonText(btn5, "L : Button %i", btn[4]);
	else SetButtonText(btn5, "L : None");
	if(btn[5])SetButtonText(btn6, "R : Button %i", btn[5]);
	else SetButtonText(btn6, "R : None");
	if(btn[6])SetButtonText(btn7, "X : Button %i", btn[6]);
	else SetButtonText(btn7, "X : None");
	if(btn[7])SetButtonText(btn8, "Y : Button %i", btn[7]);
	else SetButtonText(btn8, "Y : None");
	gtk_combo_box_append_text(GTK_COMBO_BOX(_axis_thresh), "High");
	gtk_combo_box_append_text(GTK_COMBO_BOX(_axis_thresh), "Medium");
	gtk_combo_box_append_text(GTK_COMBO_BOX(_axis_thresh), "Low");
	gtk_combo_box_set_active(GTK_COMBO_BOX(_axis_thresh), (axis_thresh/500)-1);
	//Phew !!!
	return NULL;
	#endif
}

int btnVal[8]={1<<0, 1<<1, 1<<2, 1<<3, 1<<9, 1<<8, 1<<16, 1<<17};

u32 DsPadClass::Run(u32 *value)
{
	u32 val=*value;
	//Update joystick state to detect events.
	SDL_JoystickUpdate();
	//DS POV hat.
	if(SDL_JoystickGetHat(joy, 0)&SDL_HAT_UP ||
		SDL_JoystickGetAxis(joy, 1)<=-axis_thresh)val&=~(1<<6);
	else val|=(1<<6);
	if(SDL_JoystickGetHat(joy, 0)&SDL_HAT_DOWN ||
		SDL_JoystickGetAxis(joy, 1)>=axis_thresh)val&=~(1<<7);
	else val|=(1<<7);
	if(SDL_JoystickGetHat(joy, 0)&SDL_HAT_LEFT ||
		SDL_JoystickGetAxis(joy, 0)<=-axis_thresh)val&=~(1<<5);
	else val|=(1<<5);
	if(SDL_JoystickGetHat(joy, 0)&SDL_HAT_RIGHT ||
		SDL_JoystickGetAxis(joy, 0)>=axis_thresh)val&=~(1<<4);
	else val|=(1<<4);
	//DS buttons.
	for(int i=0; i<=7; i++)
	{
		for(int j=1; j<=SDL_JoystickNumButtons(joy); j++)
		{
			if(btn[i]==j)
			{
				if(SDL_JoystickGetButton(joy, j-1))
				{
					val&=~btnVal[i];
				}
				else
				{
					val|=btnVal[i];
				}
			}
		}
	}
	*value=val;
	return 1;
}

#ifdef _WIN32

HBRUSH ReCreateSolidBrush(COLORREF color, HBRUSH old_brush)
{
	DeleteObject(old_brush);
	return CreateSolidBrush(color);
}

RECT ResetRect()
{
	RECT Rect;
	Rect.top=112;
	Rect.left=119;
	Rect.bottom=187;
	Rect.right=194;
	return Rect;
}

#endif

char btnName[8][32]={"A", "B", "Select", "Start", "L", "R", "X", "Y"};
int c_btn=0;

#ifdef _WIN32

BOOL CfgDlgProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	int i, j;
	HDC dvHdc;
	HBRUSH dvHbrush;
	RECT dvRect;
	switch(uMsg)
	{
	case WM_INITDIALOG:
		DsPad.LoadConf();
		for(i=1012; i<=1019; i++)
		{
			if(btn[i-1012])SetButtonText(GetDlgItem(hWnd, i), "%s : Button %i", btnName[i-1012], btn[i-1012]);
			else SetButtonText(GetDlgItem(hWnd, i), "%s : None", btnName[i-1012]);
		}
		AddComboItem(GetDlgItem(hWnd, IDC_AS), "High");
		AddComboItem(GetDlgItem(hWnd, IDC_AS), "Medium");
		AddComboItem(GetDlgItem(hWnd, IDC_AS), "Low");
		ComboBox_SetCurSel(GetDlgItem(hWnd, IDC_AS), (axis_thresh/500)-1);
		SetTimer(hWnd, 0x80, 10, NULL);
		return TRUE;
	case WM_COMMAND:
		switch(LOWORD(wParam))
		{
		case IDC_SC:
			DsPad.SaveConf();
			return TRUE;
		case IDC_AS:
			axis_thresh=(ComboBox_GetCurSel(GetDlgItem(hWnd, IDC_AS))+1)*500;
			DsPad.SaveConf();
			return TRUE;
		default:
			if(LOWORD(wParam)>=1012 && LOWORD(wParam)<=1019)
			{
				if(c_btn)EnableWindow(GetDlgItem(hWnd, c_btn), TRUE);
				c_btn=LOWORD(wParam);
				EnableWindow(GetDlgItem(hWnd, c_btn), FALSE);
			}
			return TRUE;
		}
	case WM_TIMER:
		COLORREF color;
		//Update joystick to detect events.
		SDL_JoystickUpdate();
		//Draw the POV hat area first.
		//Draw a beautiful frame for the POV hat area.
		//Reacts to POV & axis events.
		dvRect.top=112;
		dvRect.left=119;
		dvRect.bottom=187;
		dvRect.right=194;
		dvHbrush=CreateSolidBrush(RGB(0, 0, 0));
		dvHdc=GetDC(hWnd);
		FrameRect(dvHdc, &dvRect, dvHbrush);
		//Draw the other things into our frame.
		//To make sure the frame stays visible, we'll modify each position of 1 pixel,
		//else the frame is over-drawn.
		dvRect.top+=1;
		dvRect.left+=1;
		dvRect.bottom-=1;
		dvRect.right-=1;
		dvHbrush=ReCreateSolidBrush(RGB(224, 224, 224), dvHbrush);	//Custom func to re-create a brush easily.
		FillRect(dvHdc, &dvRect, dvHbrush);
		//Up
		dvRect.top+=2;
		dvRect.left+=32;
		dvRect.bottom=(dvRect.top+29);
		dvRect.right=(dvRect.left+8);
		dvHbrush=ReCreateSolidBrush(RGB(0, 0, 0), dvHbrush);
		FrameRect(dvHdc, &dvRect, dvHbrush);
		dvRect.top+=1;
		dvRect.left+=1;
		dvRect.bottom-=1;
		dvRect.right-=1;
		if(SDL_JoystickGetHat(joy, 0)&SDL_HAT_UP ||
			SDL_JoystickGetAxis(joy, 1)<=-axis_thresh)color=RGB(255, 0, 0);
		else color=RGB(255, 255, 255);
		dvHbrush=ReCreateSolidBrush(color, dvHbrush);
		FillRect(dvHdc, &dvRect, dvHbrush);
		//Left
		dvRect=ResetRect();
		dvRect.top+=32;
		dvRect.left+=3;
		dvRect.bottom=(dvRect.top+8);
		dvRect.right=(dvRect.left+30);
		dvHbrush=ReCreateSolidBrush(RGB(0, 0, 0), dvHbrush);
		FrameRect(dvHdc, &dvRect, dvHbrush);
		dvRect.top+=1;
		dvRect.left+=1;
		dvRect.bottom-=1;
		dvRect.right-=1;
		if(SDL_JoystickGetHat(joy, 0)&SDL_HAT_LEFT ||
			SDL_JoystickGetAxis(joy, 0)<=-axis_thresh)color=RGB(255, 0, 0);
		else color=RGB(255, 255, 255);
		dvHbrush=ReCreateSolidBrush(color, dvHbrush);
		FillRect(dvHdc, &dvRect, dvHbrush);
		//Right
		dvRect=ResetRect();
		dvRect.top+=32;
		dvRect.left+=41;
		dvRect.bottom=(dvRect.top+8);
		dvRect.right=(dvRect.left+31);
		dvHbrush=ReCreateSolidBrush(RGB(0, 0, 0), dvHbrush);
		FrameRect(dvHdc, &dvRect, dvHbrush);
		dvRect.top+=1;
		dvRect.left+=1;
		dvRect.bottom-=1;
		dvRect.right-=1;
		if(SDL_JoystickGetHat(joy, 0)&SDL_HAT_RIGHT ||
			SDL_JoystickGetAxis(joy, 0)>=axis_thresh)color=RGB(255, 0, 0);
		else color=RGB(255, 255, 255);
		dvHbrush=ReCreateSolidBrush(color, dvHbrush);
		FillRect(dvHdc, &dvRect, dvHbrush);
		//Down
		dvRect=ResetRect();
		dvRect.top+=40;
		dvRect.left+=33;
		dvRect.bottom=(dvRect.top+32);
		dvRect.right=(dvRect.left+8);
		dvHbrush=ReCreateSolidBrush(RGB(0, 0, 0), dvHbrush);
		FrameRect(dvHdc, &dvRect, dvHbrush);
		dvRect.top+=1;
		dvRect.left+=1;
		dvRect.bottom-=1;
		dvRect.right-=1;
		if(SDL_JoystickGetHat(joy, 0)&SDL_HAT_DOWN ||
			SDL_JoystickGetAxis(joy, 1)>=axis_thresh)color=RGB(255, 0, 0);
		else color=RGB(255, 255, 255);
		dvHbrush=ReCreateSolidBrush(color, dvHbrush);
		FillRect(dvHdc, &dvRect, dvHbrush);
		//Center
		dvRect=ResetRect();
		dvRect.top+=31;
		dvRect.left+=32;
		dvRect.bottom=(dvRect.top+10);
		dvRect.right=(dvRect.left+10);
		dvHbrush=ReCreateSolidBrush(RGB(0, 0, 0), dvHbrush);
		FrameRect(dvHdc, &dvRect, dvHbrush);
		dvRect.top+=1;
		dvRect.left+=1;
		dvRect.bottom-=1;
		dvRect.right-=1;
		if(SDL_JoystickGetHat(joy, 0)==SDL_HAT_CENTERED && 
			(SDL_JoystickGetAxis(joy, 0)<=axis_thresh && SDL_JoystickGetAxis(joy, 0)>=-axis_thresh) && 
			(SDL_JoystickGetAxis(joy, 1)<=axis_thresh && SDL_JoystickGetAxis(joy, 1)>=-axis_thresh))color=RGB(255, 0, 0);
		else color=RGB(255, 255, 255);
		dvHbrush=ReCreateSolidBrush(color, dvHbrush);
		FillRect(dvHdc, &dvRect, dvHbrush);
		ReleaseDC(hWnd, dvHdc);
		DeleteObject(dvHbrush);
		//Now detect button events.
		if(c_btn)
		{
			for(j=1; j<=SDL_JoystickNumButtons(joy); j++)
			{
				if(SDL_JoystickGetButton(joy, j-1))
				{
					btn[c_btn-1012]=j;
					EnableWindow(GetDlgItem(hWnd, c_btn), TRUE);
					SetButtonText(GetDlgItem(hWnd, c_btn), "%s : Button %i", btnName[c_btn-1012], j);
					DsPad.SaveConf();
					c_btn=0;
				}
			}
		}
		return TRUE;
	}
	return FALSE;
}

#else

void on_btn_clicked(GtkButton *button){
	int index=atoi(&(*(gtk_widget_get_name(GTK_WIDGET(button))))+3);
	for(;;){
		SDL_JoystickUpdate();
		for(int i=0; i<SDL_JoystickNumButtons(joy); i++){
			if(SDL_JoystickGetButton(joy, i)){
				btn[index-1]=i+1;
				SetButtonText(button, "%s : Button %i", btnName[index-1], i+1);
				DsPad.SaveConf();
				return;
			}
		}
	}
}

void on_axisthresh_changed(GtkComboBox *combo){
	axis_thresh=(gtk_combo_box_get_active(combo)+1)*500;
	DsPad.SaveConf();
}

void on_saveconf_clicked(){
	DsPad.SaveConf();
}

#endif
